@@ -156,9 +156,10 @@ adw_status_page_dispose (GObject *object)
 
   adw_status_page_set_child (self, NULL);
 
-  gtk_widget_unparent (self->scrolled_window);
   self->user_widget = NULL;
 
+  gtk_widget_dispose_template (GTK_WIDGET (self), ADW_TYPE_STATUS_PAGE);
+
   G_OBJECT_CLASS (adw_status_page_parent_class)->dispose (object);
 }
 
@@ -214,6 +215,8 @@ adw_status_page_class_init (AdwStatusPageClass *klass)
    * AdwStatusPage:title: (attributes org.gtk.Property.get=adw_status_page_get_title org.gtk.Property.set=adw_status_page_set_title)
    *
    * The title to be displayed below the icon.
+   * 
+   * It is not parsed as Pango markup.
    */
   props[PROP_TITLE] =
     g_param_spec_string ("title", NULL, NULL,
@@ -223,7 +226,7 @@ adw_status_page_class_init (AdwStatusPageClass *klass)
   /**
    * AdwStatusPage:description: (attributes org.gtk.Property.get=adw_status_page_get_description org.gtk.Property.set=adw_status_page_set_description)
    *
-   * The description to be displayed below the title.
+   * The description markup to be displayed below the title.
    */
   props[PROP_DESCRIPTION] =
     g_param_spec_string ("description", NULL, NULL,
@@ -342,8 +345,7 @@ adw_status_page_set_icon_name (AdwStatusPage *self,
     g_object_notify_by_pspec (G_OBJECT (self), props[PROP_PAINTABLE]);
   }
 
-  g_free (self->icon_name);
-  self->icon_name = g_strdup (icon_name);
+  g_set_str (&self->icon_name, icon_name);
   gtk_image_set_from_icon_name (self->image, self->icon_name);
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_ICON_NAME]);
 
@@ -422,7 +424,7 @@ adw_status_page_get_title (AdwStatusPage *self)
  *
  * Sets the title for @self.
  *
- * The title is displayed below the icon.
+ * The title is displayed below the icon. It is not parsed as Pango markup.
  */
 void
 adw_status_page_set_title (AdwStatusPage *self,
@@ -442,7 +444,7 @@ adw_status_page_set_title (AdwStatusPage *self,
  * adw_status_page_get_description: (attributes org.gtk.Method.get_property=description)
  * @self: a status page
  *
- * Gets the description for @self.
+ * Gets the description markup for @self.
  *
  * Returns: (nullable): the description
  */
@@ -459,9 +461,9 @@ adw_status_page_get_description (AdwStatusPage *self)
  * @self: a status page
  * @description: (nullable): the description
  *
- * Sets the description for @self.
+ * Sets the description markup for @self.
  *
- * The description is displayed below the title.
+ * The description is displayed below the title. It is parsed as Pango markup.
  */
 void
 adw_status_page_set_description (AdwStatusPage *self,
@@ -505,6 +507,10 @@ adw_status_page_set_child (AdwStatusPage *self,
                            GtkWidget     *child)
 {
   g_return_if_fail (ADW_IS_STATUS_PAGE (self));
+  g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
+
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
 
   if (child == self->user_widget)
     return;
